@@ -1,4 +1,4 @@
-/*      $Id: events.c 28239 2008-10-14 08:04:11Z olivier $
+/*      $Id: events.c 29542 2009-02-23 10:06:52Z olivier $
 
         This program is free software; you can redistribute it and/or modify
         it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
 
 
         oroborus - (c) 2001 Ken Lynch
-        xfwm4    - (c) 2002-2008 Olivier Fourdan
+        xfwm4    - (c) 2002-2009 Olivier Fourdan
 
  */
 
@@ -50,6 +50,7 @@
 #include "client.h"
 #include "moveresize.h"
 #include "cycle.h"
+#include "placement.h"
 #include "stacking.h"
 #include "transients.h"
 #include "focus.h"
@@ -77,13 +78,6 @@
                                  ButtonPressMask | \
                                  ButtonReleaseMask)
 
-#define MODIFIER_MASK           (ShiftMask | \
-                                 ControlMask | \
-                                 AltMask | \
-                                 MetaMask | \
-                                 SuperMask | \
-                                 HyperMask)
-
 static GdkAtom atom_rcfiles = GDK_NONE;
 static xfwmWindow menu_event_window;
 static int edge_scroll_x = 0;
@@ -262,7 +256,7 @@ typeOfClick (ScreenInfo *screen_info, Window w, XEvent * ev, gboolean allow_doub
 static gboolean
 check_button_time (XButtonEvent *ev)
 {
-    static Time last_button_time = (Time) CurrentTime;
+    static guint32 last_button_time = CurrentTime;
 
     if (last_button_time > ev->time)
     {
@@ -291,24 +285,6 @@ handleMotionNotify (DisplayInfo *display_info, XMotionEvent * ev)
     return EVENT_FILTER_REMOVE;
 }
 
-static int
-getKeyPressed (ScreenInfo *screen_info, XKeyEvent * ev)
-{
-    int key, state;
-
-    state = ev->state & MODIFIER_MASK;
-    for (key = 0; key < KEY_COUNT; key++)
-    {
-        if ((screen_info->params->keys[key].keycode == ev->keycode)
-            && (screen_info->params->keys[key].modifier == state))
-        {
-            break;
-        }
-    }
-
-    return key;
-}
-
 static eventFilterStatus
 handleKeyPress (DisplayInfo *display_info, XKeyEvent * ev)
 {
@@ -321,11 +297,10 @@ handleKeyPress (DisplayInfo *display_info, XKeyEvent * ev)
     TRACE ("entering handleKeyEvent");
 
     ev_screen_info = myDisplayGetScreenFromRoot (display_info, ev->root);
+
     if (!ev_screen_info)
     {
-        /* Release queued events */
-        XAllowEvents (display_info->dpy, SyncKeyboard, CurrentTime);
-
+        XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
         return EVENT_FILTER_PASS;
     }
 
@@ -334,76 +309,97 @@ handleKeyPress (DisplayInfo *display_info, XKeyEvent * ev)
     if (c)
     {
         screen_info = c->screen_info;
-        key = getKeyPressed (screen_info, ev);
+        key = myScreenGetKeyPressed (screen_info, ev);
         status = EVENT_FILTER_REMOVE;
 
         switch (key)
         {
             case KEY_MOVE:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 clientMove (c, (XEvent *) ev);
                 break;
             case KEY_RESIZE:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 clientResize (c, CORNER_BOTTOM_RIGHT, (XEvent *) ev);
                 break;
             case KEY_CYCLE_WINDOWS:
+            case KEY_CYCLE_REVERSE_WINDOWS:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 clientCycle (c, ev);
                 break;
             case KEY_CLOSE_WINDOW:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 clientClose (c);
                 break;
             case KEY_HIDE_WINDOW:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 if (CLIENT_CAN_HIDE_WINDOW (c))
                 {
                     clientWithdraw (c, c->win_workspace, TRUE);
                 }
                 break;
             case KEY_MAXIMIZE_WINDOW:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 clientToggleMaximized (c, WIN_STATE_MAXIMIZED, TRUE);
                 break;
             case KEY_MAXIMIZE_VERT:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 clientToggleMaximized (c, WIN_STATE_MAXIMIZED_VERT, TRUE);
                 break;
             case KEY_MAXIMIZE_HORIZ:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 clientToggleMaximized (c, WIN_STATE_MAXIMIZED_HORIZ, TRUE);
                 break;
             case KEY_SHADE_WINDOW:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 clientToggleShaded (c);
                 break;
             case KEY_STICK_WINDOW:
-                if (CLIENT_CAN_STICK_WINDOW(c))
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
+                if (FLAG_TEST(c->xfwm_flags, XFWM_FLAG_HAS_STICK))
                 {
                     clientToggleSticky (c, TRUE);
                     frameQueueDraw (c, FALSE);
                 }
                 break;
             case KEY_RAISE_WINDOW:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 clientRaise (c, None);
                 break;
             case KEY_LOWER_WINDOW:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 clientLower (c, None);
                 break;
             case KEY_TOGGLE_ABOVE:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 clientToggleLayerAbove (c);
                 break;
             case KEY_TOGGLE_FULLSCREEN:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 clientToggleFullscreen (c);
                 break;
             case KEY_MOVE_NEXT_WORKSPACE:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 workspaceSwitch (screen_info, screen_info->current_ws + 1, c, TRUE, ev->time);
                 break;
             case KEY_MOVE_PREV_WORKSPACE:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 workspaceSwitch (screen_info, screen_info->current_ws - 1, c, TRUE, ev->time);
                 break;
             case KEY_MOVE_UP_WORKSPACE:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 workspaceMove (screen_info, -1, 0, c, ev->time);
                 break;
             case KEY_MOVE_DOWN_WORKSPACE:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 workspaceMove (screen_info, 1, 0, c, ev->time);
                 break;
             case KEY_MOVE_LEFT_WORKSPACE:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 workspaceMove (screen_info, 0, -1, c, ev->time);
                 break;
             case KEY_MOVE_RIGHT_WORKSPACE:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 workspaceMove (screen_info, 0, 1, c, ev->time);
                 break;
             case KEY_MOVE_WORKSPACE_1:
@@ -418,6 +414,7 @@ handleKeyPress (DisplayInfo *display_info, XKeyEvent * ev)
             case KEY_MOVE_WORKSPACE_10:
             case KEY_MOVE_WORKSPACE_11:
             case KEY_MOVE_WORKSPACE_12:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 if (key - KEY_MOVE_WORKSPACE_1 < screen_info->workspace_count)
                 {
                     clientRaise (c, None);
@@ -425,18 +422,21 @@ handleKeyPress (DisplayInfo *display_info, XKeyEvent * ev)
                 }
                 break;
             case KEY_POPUP_MENU:
-                XAllowEvents (display_info->dpy, SyncKeyboard, CurrentTime);
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 show_window_menu (c, frameX (c) + frameLeft (c),
                                      frameY (c) + frameTop (c),
-                                     Button1, GDK_CURRENT_TIME);
+                                     Button1, ev->time);
                 break;
             case KEY_FILL_WINDOW:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 clientFill (c, CLIENT_FILL);
                 break;
             case KEY_FILL_VERT:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 clientFill (c, CLIENT_FILL_VERT);
                 break;
             case KEY_FILL_HORIZ:
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 clientFill (c, CLIENT_FILL_HORIZ);
                 break;
             default:
@@ -445,11 +445,12 @@ handleKeyPress (DisplayInfo *display_info, XKeyEvent * ev)
     }
     else
     {
-        key = getKeyPressed (ev_screen_info, ev);
+        key = myScreenGetKeyPressed (ev_screen_info, ev);
         switch (key)
         {
             case KEY_CYCLE_WINDOWS:
                 status = EVENT_FILTER_REMOVE;
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 if (ev_screen_info->clients)
                 {
                     clientCycle (ev_screen_info->clients->prev, ev);
@@ -457,6 +458,7 @@ handleKeyPress (DisplayInfo *display_info, XKeyEvent * ev)
                 break;
             case KEY_CLOSE_WINDOW:
                 status = EVENT_FILTER_REMOVE;
+                XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
                 if (display_info->session)
                 {
                     logout_session (display_info->session);
@@ -471,40 +473,50 @@ handleKeyPress (DisplayInfo *display_info, XKeyEvent * ev)
     {
         case KEY_NEXT_WORKSPACE:
             status = EVENT_FILTER_REMOVE;
+            XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
             workspaceSwitch (ev_screen_info, ev_screen_info->current_ws + 1, NULL, TRUE, ev->time);
             break;
         case KEY_PREV_WORKSPACE:
             status = EVENT_FILTER_REMOVE;
+            XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
             workspaceSwitch (ev_screen_info, ev_screen_info->current_ws - 1, NULL, TRUE, ev->time);
             break;
         case KEY_UP_WORKSPACE:
             status = EVENT_FILTER_REMOVE;
+            XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
             workspaceMove(ev_screen_info, -1, 0, NULL, ev->time);
             break;
         case KEY_DOWN_WORKSPACE:
             status = EVENT_FILTER_REMOVE;
+            XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
             workspaceMove(ev_screen_info, 1, 0, NULL, ev->time);
             break;
         case KEY_LEFT_WORKSPACE:
             status = EVENT_FILTER_REMOVE;
+            XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
             workspaceMove(ev_screen_info, 0, -1, NULL, ev->time);
             break;
         case KEY_RIGHT_WORKSPACE:
             status = EVENT_FILTER_REMOVE;
+            XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
             workspaceMove(ev_screen_info, 0, 1, NULL, ev->time);
             break;
         case KEY_ADD_WORKSPACE:
             status = EVENT_FILTER_REMOVE;
+            XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
             workspaceSetCount (ev_screen_info, ev_screen_info->workspace_count + 1);
             break;
         case KEY_DEL_WORKSPACE:
             status = EVENT_FILTER_REMOVE;
+            XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
             workspaceSetCount (ev_screen_info, ev_screen_info->workspace_count - 1);
             break;
         case KEY_ADD_ADJACENT_WORKSPACE:
+            XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
             workspaceInsert (ev_screen_info, ev_screen_info->current_ws + 1);
             break;
         case KEY_DEL_ACTIVE_WORKSPACE:
+            XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
             workspaceDelete (ev_screen_info, ev_screen_info->current_ws);
             break;
         case KEY_WORKSPACE_1:
@@ -520,6 +532,7 @@ handleKeyPress (DisplayInfo *display_info, XKeyEvent * ev)
         case KEY_WORKSPACE_11:
         case KEY_WORKSPACE_12:
             status = EVENT_FILTER_REMOVE;
+            XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
             if (key - KEY_WORKSPACE_1 < ev_screen_info->workspace_count)
             {
                 workspaceSwitch (ev_screen_info, key - KEY_WORKSPACE_1, NULL, TRUE, ev->time);
@@ -527,15 +540,13 @@ handleKeyPress (DisplayInfo *display_info, XKeyEvent * ev)
             break;
         case KEY_SHOW_DESKTOP:
             status = EVENT_FILTER_REMOVE;
+            XAllowEvents (display_info->dpy, AsyncKeyboard, ev->time);
             toggle_show_desktop (ev_screen_info);
             break;
         default:
             break;
     }
 
-    /* Release queued events */
-    XAllowEvents (display_info->dpy, SyncKeyboard, CurrentTime);
-
     return status;
 }
 
@@ -544,9 +555,6 @@ handleKeyRelease (DisplayInfo *display_info, XKeyEvent * ev)
 {
     TRACE ("entering handleKeyRelease");
 
-    /* Release queued events */
-    XAllowEvents (display_info->dpy, SyncKeyboard, CurrentTime);
-
     return EVENT_FILTER_PASS;
 }
 
@@ -816,7 +824,7 @@ titleButton (Client * c, int state, XButtonEvent * ev)
 static void
 rootScrollButton (DisplayInfo *display_info, XButtonEvent * ev)
 {
-    static Time lastscroll = (Time) CurrentTime;
+    static guint32 lastscroll = CurrentTime;
     ScreenInfo *screen_info;
 
     if ((ev->time - lastscroll) < 25)  /* ms */
@@ -1209,12 +1217,14 @@ handleUnmapNotify (DisplayInfo *display_info, XUnmapEvent * ev)
          */
         if ((ev->event == screen_info->xroot) && (ev->send_event))
         {
-            TRACE ("ICCCM UnmapNotify for \"%s\"", c->name);
-            list_of_windows = clientListTransientOrModal (c);
-            clientPassFocus (screen_info, c, list_of_windows);
-            clientUnframe (c, FALSE);
-            g_list_free (list_of_windows);
-
+            if (!FLAG_TEST (c->xfwm_flags, XFWM_FLAG_VISIBLE))
+            {
+                TRACE ("ICCCM UnmapNotify for \"%s\"", c->name);
+                list_of_windows = clientListTransientOrModal (c);
+                clientPassFocus (screen_info, c, list_of_windows);
+                clientUnframe (c, FALSE);
+                g_list_free (list_of_windows);
+            }
             return status;
         }
 
@@ -1247,7 +1257,7 @@ handleConfigureNotify (DisplayInfo *display_info, XConfigureEvent * ev)
 static eventFilterStatus
 handleConfigureRequest (DisplayInfo *display_info, XConfigureRequestEvent * ev)
 {
-    Client *c = NULL;
+    Client *c;
     XWindowChanges wc;
 
     TRACE ("entering handleConfigureRequest");
@@ -1293,105 +1303,14 @@ handleConfigureRequest (DisplayInfo *display_info, XConfigureRequestEvent * ev)
     }
     if (c)
     {
-        gboolean constrained = FALSE;
-        ScreenInfo *screen_info = c->screen_info;
-
         TRACE ("handleConfigureRequest managed window \"%s\" (0x%lx)", c->name, c->window);
         if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_MOVING_RESIZING))
         {
             /* Sorry, but it's not the right time for configure request */
             return EVENT_FILTER_REMOVE;
         }
-        if (c->type == WINDOW_DESKTOP)
-        {
-            /* Ignore stacking request for DESKTOP windows */
-            ev->value_mask &= ~(CWSibling | CWStackMode);
-        }
-        clientAdjustCoordGravity (c, &ev->value_mask, &wc);
-
-        if (FLAG_TEST (c->flags, CLIENT_FLAG_FULLSCREEN))
-        {
-            GdkRectangle rect;
-            gint monitor_nbr;
-            int cx, cy;
-
-            /* size request from fullscreen windows get fullscreen */
-
-            cx = frameX (c) + (frameWidth (c) / 2);
-            cy = frameY (c) + (frameHeight (c) / 2);
-
-            monitor_nbr = find_monitor_at_point (screen_info->gscr, cx, cy);
-            gdk_screen_get_monitor_geometry (screen_info->gscr, monitor_nbr, &rect);
-
-            wc.x = rect.x;
-            wc.y = rect.y;
-            wc.width = rect.width;
-            wc.height = rect.height;
-
-            ev->value_mask |= (CWX | CWY | CWWidth | CWHeight);
-        }
-        else if (FLAG_TEST_ALL (c->flags, CLIENT_FLAG_MAXIMIZED)
-                 && (screen_info->params->borderless_maximize))
-        {
-            wc.x = c->x;
-            wc.y = c->y;
-            wc.width = c->width;
-            wc.height = c->height;
-        }
-        /* Clean up buggy requests that set all flags */
-        if ((ev->value_mask & CWX) && (wc.x == c->x))
-        {
-            ev->value_mask &= ~CWX;
-        }
-        if ((ev->value_mask & CWY) && (wc.y == c->y))
-        {
-            ev->value_mask &= ~CWY;
-        }
-        if ((ev->value_mask & CWWidth) && (wc.width == c->width))
-        {
-            ev->value_mask &= ~CWWidth;
-        }
-        if ((ev->value_mask & CWHeight) && (wc.height == c->height))
-        {
-            ev->value_mask &= ~CWHeight;
-        }
-        /* Still a move/resize after cleanup? */
-        if (ev->value_mask & (CWX | CWY | CWWidth | CWHeight))
-        {
-            if (FLAG_TEST (c->flags, CLIENT_FLAG_MAXIMIZED))
-            {
-                clientRemoveMaximizeFlag (c);
-            }
-            constrained = TRUE;
-        }
-
-        /*
-           Let's say that if the client performs a XRaiseWindow, we show the window if focus
-           stealing prevention is not activated, otherwise we just set the "demands attention"
-           flag...
-         */
-        if ((ev->value_mask & CWStackMode) && (wc.stack_mode == Above) && (wc.sibling == None) && !(c->type & WINDOW_TYPE_DONT_FOCUS))
-        {
-            Client *last_raised;
-
-            last_raised = clientGetLastRaise (screen_info);
-            if (last_raised && (c != last_raised))
-            {
-                if ((screen_info->params->prevent_focus_stealing) && (screen_info->params->activate_action == ACTIVATE_ACTION_NONE))
-                {
-                    ev->value_mask &= ~(CWSibling | CWStackMode);
-                    TRACE ("Setting WM_STATE_DEMANDS_ATTENTION flag on \"%s\" (0x%lx)", c->name, c->window);
-                    FLAG_SET (c->flags, CLIENT_FLAG_DEMANDS_ATTENTION);
-                    clientSetNetState (c);
-                }
-                else
-                {
-                    clientActivate (c, getXServerTime (display_info));
-                }
-            }
-        }
-
-        clientConfigure (c, &wc, ev->value_mask, (constrained ? CFG_CONSTRAINED : 0) | CFG_REQUEST);
+        clientAdjustCoordGravity (c, c->gravity, &ev->value_mask, &wc);
+        clientMoveResizeWindow (c, &wc, ev->value_mask);
     }
     else
     {
@@ -1405,7 +1324,7 @@ handleConfigureRequest (DisplayInfo *display_info, XConfigureRequestEvent * ev)
 static eventFilterStatus
 handleEnterNotify (DisplayInfo *display_info, XCrossingEvent * ev)
 {
-    static Time lastresist = (Time) CurrentTime;
+    static guint32 lastresist = CurrentTime;
     ScreenInfo *screen_info;
     Client *c;
     int b;
@@ -1494,7 +1413,7 @@ handleEnterNotify (DisplayInfo *display_info, XCrossingEvent * ev)
         && screen_info->params->wrap_resistance)
     {
         int msx, msy, maxx, maxy;
-        int rx, ry;
+        int rx, ry, delta;
 
         msx = ev->x_root;
         msy = ev->y_root;
@@ -1550,18 +1469,19 @@ handleEnterNotify (DisplayInfo *display_info, XCrossingEvent * ev)
         if (edge_scroll_x > screen_info->params->wrap_resistance)
         {
             edge_scroll_x = 0;
+            delta = MAX (9 * maxx / 10, maxx - 5 * screen_info->params->wrap_resistance);
             if (msx == 0)
             {
                 if (workspaceMove (screen_info, 0, -1, NULL, ev->time))
                 {
-                    rx = 4 * maxx / 5;
+                    rx = delta;
                 }
             }
             else
             {
                 if (workspaceMove (screen_info, 0, 1, NULL, ev->time))
                 {
-                    rx = -4 * maxx / 5;
+                    rx = -delta;
                 }
             }
             warp_pointer = TRUE;
@@ -1569,18 +1489,19 @@ handleEnterNotify (DisplayInfo *display_info, XCrossingEvent * ev)
         if (edge_scroll_y > screen_info->params->wrap_resistance)
         {
             edge_scroll_y = 0;
+            delta = MAX (9 * maxy / 10, maxy - 5 * screen_info->params->wrap_resistance);
             if (msy == 0)
             {
                 if (workspaceMove (screen_info, -1, 0, NULL, ev->time))
                 {
-                    ry = 4 * maxy / 5;
+                    ry = delta;
                 }
             }
             else
             {
                 if (workspaceMove (screen_info, 1, 0, NULL, ev->time))
                 {
-                    ry = -4 * maxy / 5;
+                    ry = -delta;
                 }
             }
             warp_pointer = TRUE;
@@ -1911,11 +1832,7 @@ handlePropertyNotify (DisplayInfo *display_info, XPropertyEvent * ev)
         else if (ev->atom == display_info->atoms[NET_WM_USER_TIME])
         {
             TRACE ("client \"%s\" (0x%lx) has received a NET_WM_USER_TIME notify", c->name, c->window);
-            if (getHint (display_info, c->window, NET_WM_USER_TIME, (long *) &c->user_time) && (c->user_time != 0))
-            {
-                myDisplaySetLastUserTime (display_info, c->user_time);
-                FLAG_SET (c->flags, CLIENT_FLAG_HAS_USER_TIME);
-            }
+            clientGetUserTime (c);
         }
         else if (ev->atom == display_info->atoms[NET_WM_USER_TIME_WINDOW])
         {
@@ -2020,7 +1937,6 @@ handleClientMessage (DisplayInfo *display_info, XClientMessageEvent * ev)
     eventFilterStatus status;
     ScreenInfo *screen_info;
     Client *c;
-    gboolean is_transient;
 
     TRACE ("entering handleClientMessage");
 
@@ -2030,7 +1946,6 @@ handleClientMessage (DisplayInfo *display_info, XClientMessageEvent * ev)
     {
         status = EVENT_FILTER_REMOVE;
         screen_info = c->screen_info;
-        is_transient = clientIsValidTransientOrModal (c);
 
         if ((ev->message_type == display_info->atoms[WM_CHANGE_STATE]) && (ev->format == 32) && (ev->data.l[0] == IconicState))
         {
@@ -2048,7 +1963,7 @@ handleClientMessage (DisplayInfo *display_info, XClientMessageEvent * ev)
         else if ((ev->message_type == display_info->atoms[WIN_LAYER]) && (ev->format == 32))
         {
             TRACE ("client \"%s\" (0x%lx) has received a WIN_LAYER event", c->name, c->window);
-            if ((ev->data.l[0] != c->win_layer) && !is_transient)
+            if (ev->data.l[0] != c->win_layer)
             {
                 clientSetLayer (c, ev->data.l[0]);
             }
@@ -2056,7 +1971,7 @@ handleClientMessage (DisplayInfo *display_info, XClientMessageEvent * ev)
         else if ((ev->message_type == display_info->atoms[WIN_WORKSPACE]) && (ev->format == 32))
         {
             TRACE ("client \"%s\" (0x%lx) has received a WIN_WORKSPACE event", c->name, c->window);
-            if ((ev->data.l[0] != c->win_workspace) && !is_transient)
+            if (ev->data.l[0] != c->win_workspace)
             {
                 clientSetWorkspace (c, ev->data.l[0], TRUE);
             }
@@ -2064,27 +1979,24 @@ handleClientMessage (DisplayInfo *display_info, XClientMessageEvent * ev)
         else if ((ev->message_type == display_info->atoms[NET_WM_DESKTOP]) && (ev->format == 32))
         {
             TRACE ("client \"%s\" (0x%lx) has received a NET_WM_DESKTOP event", c->name, c->window);
-            if (!is_transient)
+            if (ev->data.l[0] == ALL_WORKSPACES)
             {
-                if (ev->data.l[0] == ALL_WORKSPACES)
+                if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_HAS_STICK) && !FLAG_TEST (c->flags, CLIENT_FLAG_STICKY))
                 {
-                    if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_HAS_STICK) && !FLAG_TEST (c->flags, CLIENT_FLAG_STICKY))
-                    {
-                        clientStick (c, TRUE);
-                        frameQueueDraw (c, FALSE);
-                    }
+                    clientStick (c, TRUE);
+                    frameQueueDraw (c, FALSE);
                 }
-                else
+            }
+            else
+            {
+                if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_HAS_STICK) && FLAG_TEST (c->flags, CLIENT_FLAG_STICKY))
                 {
-                    if (FLAG_TEST (c->xfwm_flags, XFWM_FLAG_HAS_STICK) && FLAG_TEST (c->flags, CLIENT_FLAG_STICKY))
-                    {
-                        clientUnstick (c, TRUE);
-                        frameQueueDraw (c, FALSE);
-                    }
-                    if (ev->data.l[0] != c->win_workspace)
-                    {
-                        clientSetWorkspace (c, ev->data.l[0], TRUE);
-                    }
+                    clientUnstick (c, TRUE);
+                    frameQueueDraw (c, FALSE);
+                }
+                if (ev->data.l[0] != c->win_workspace)
+                {
+                    clientSetWorkspace (c, ev->data.l[0], TRUE);
                 }
             }
         }
@@ -2103,20 +2015,25 @@ handleClientMessage (DisplayInfo *display_info, XClientMessageEvent * ev)
             TRACE ("client \"%s\" (0x%lx) has received a NET_WM_MOVERESIZE event", c->name, c->window);
             clientNetMoveResize (c, ev);
         }
+        else if ((ev->message_type == display_info->atoms[NET_MOVERESIZE_WINDOW]) && (ev->format == 32))
+        {
+            TRACE ("client \"%s\" (0x%lx) has received a NET_MOVERESIZE_WINDOW event", c->name, c->window);
+            clientNetMoveResizeWindow (c, ev);
+        }
         else if ((ev->message_type == display_info->atoms[NET_ACTIVE_WINDOW]) && (ev->format == 32))
         {
             gboolean source_is_application;
-            Time ev_time;
+            guint32 ev_time;
 
-            ev_time = myDisplayGetTime (display_info, (Time) ev->data.l[1]);
+            ev_time = myDisplayGetTime (display_info, (guint32) ev->data.l[1]);
             source_is_application = (ev->data.l[0] == 1);
 
             TRACE ("client \"%s\" (0x%lx) has received a NET_ACTIVE_WINDOW event", c->name, c->window);
             if (source_is_application)
             {
-                Time current = myDisplayGetLastUserTime (display_info);
+                guint32 current = myDisplayGetLastUserTime (display_info);
 
-                TRACE ("Time of event received is %u, current XServer time is %u", (unsigned int) ev_time, (unsigned int) current);
+                TRACE ("Time of event received is %u, current XServer time is %u", (guint32) ev_time, (guint32) current);
                 if ((screen_info->params->prevent_focus_stealing) && TIMESTAMP_IS_BEFORE(ev_time, current))
                 {
                     TRACE ("Setting WM_STATE_DEMANDS_ATTENTION flag on \"%s\" (0x%lx)", c->name, c->window);
@@ -2140,6 +2057,12 @@ handleClientMessage (DisplayInfo *display_info, XClientMessageEvent * ev)
             setNetFrameExtents (display_info, c->window, frameTop (c), frameLeft (c),
                                                          frameRight (c), frameBottom (c));
         }
+        else if (ev->message_type == display_info->atoms[NET_WM_FULLSCREEN_MONITORS])
+        {
+            TRACE ("client \"%s\" (0x%lx) has received a NET_WM_FULLSCREEN_MONITORS event", c->name, c->window);
+            clientSetFullscreenMonitor (c, (gint) ev->data.l[0], (gint) ev->data.l[1],
+                                           (gint) ev->data.l[2], (gint) ev->data.l[3]);
+        }
     }
     else
     {
@@ -2158,7 +2081,7 @@ handleClientMessage (DisplayInfo *display_info, XClientMessageEvent * ev)
                 (ev->data.l[0] != screen_info->current_ws))
             {
                 workspaceSwitch (screen_info, ev->data.l[0], NULL, TRUE,
-                                 myDisplayGetTime (display_info, (Time) ev->data.l[1]));
+                                 myDisplayGetTime (display_info, (guint32) ev->data.l[1]));
             }
         }
         else if (((ev->message_type == display_info->atoms[WIN_WORKSPACE_COUNT]) ||
@@ -2214,7 +2137,7 @@ handleClientMessage (DisplayInfo *display_info, XClientMessageEvent * ev)
             if (ev->data.l[0] == display_info->atoms[NET_WM_PING])
             {
                 TRACE ("root has received a NET_WM_PING (pong) event\n");
-                clientReceiveNetWMPong (screen_info, (Time) ev->data.l[1]);
+                clientReceiveNetWMPong (screen_info, (guint32) ev->data.l[1]);
             }
         }
         else
@@ -2306,27 +2229,9 @@ handleColormapNotify (DisplayInfo *display_info, XColormapEvent * ev)
 }
 
 static eventFilterStatus
-handleMappingNotify (DisplayInfo *display_info, XMappingEvent * ev)
+handleReparentNotify (DisplayInfo *display_info, XReparentEvent * ev)
 {
-    TRACE ("entering handleMappingNotify");
-
-    /* Refreshes the stored modifier and keymap information */
-    XRefreshKeyboardMapping (ev);
-
-    /* Update internal modifiers masks if necessary */
-    if (ev->request == MappingModifier)
-    {
-        TRACE ("handleMappingNotify: modifiers mapping has changed");
-        initModifiers (display_info->dpy);
-    }
-
-    /* Regrab all keys if the notify is for keyboard (ie not pointer) */
-    if (ev->request != MappingPointer)
-    {
-        TRACE ("handleMappingNotify: Reload settings");
-        reloadSettings (display_info, UPDATE_BUTTON_GRABS);
-    }
-
+    TRACE ("entering handleReparentNotify, 0x%lx reparented in 0x%lx", ev->window, ev->parent);
     return EVENT_FILTER_PASS;
 }
 
@@ -2425,8 +2330,8 @@ handleEvent (DisplayInfo *display_info, XEvent * ev)
         case ColormapNotify:
             handleColormapNotify (display_info, (XColormapEvent *) ev);
             break;
-        case MappingNotify:
-            status = handleMappingNotify (display_info, (XMappingEvent *) ev);
+        case ReparentNotify:
+            status = handleReparentNotify (display_info, (XReparentEvent *) ev);
             break;
         default:
             if ((display_info->have_shape) && (ev->type == display_info->shape_event_base))
@@ -2449,6 +2354,11 @@ handleEvent (DisplayInfo *display_info, XEvent * ev)
         }
         else if (display_info->quit)
         {
+            /*
+             * Qutting on purpose, update session manager so
+             * it does not restart the program immediately
+             */
+            client_session_set_restart_style(display_info->session, SESSION_RESTART_IF_RUNNING);
             gtk_main_quit ();
         }
     }
@@ -2585,6 +2495,7 @@ show_window_menu (Client *c, gint px, gint py, guint button, guint32 time)
     Menu *menu;
     MenuOp ops;
     MenuOp insensitive;
+    gboolean is_transient;
     gint x, y;
 
     TRACE ("entering show_window_menu");
@@ -2601,6 +2512,7 @@ show_window_menu (Client *c, gint px, gint py, guint button, guint32 time)
 
     screen_info = c->screen_info;
     display_info = screen_info->display_info;
+    is_transient = clientIsValidTransientOrModal (c);
 
     x = px;
     y = py;
@@ -2657,7 +2569,7 @@ show_window_menu (Client *c, gint px, gint py, guint button, guint32 time)
         insensitive |= MENU_OP_DELETE;
     }
 
-    if (!CLIENT_CAN_STICK_WINDOW(c))
+    if (is_transient || !FLAG_TEST(c->xfwm_flags, XFWM_FLAG_HAS_STICK))
     {
         insensitive |= MENU_OP_STICK | MENU_OP_UNSTICK;
     }
@@ -2697,7 +2609,7 @@ show_window_menu (Client *c, gint px, gint py, guint button, guint32 time)
         ops |= MENU_OP_FULLSCREEN;
     }
 
-    if (clientIsTransientOrModal (c) || (c->type != WINDOW_NORMAL))
+    if (is_transient || (c->type != WINDOW_NORMAL))
     {
         insensitive |= MENU_OP_FULLSCREEN | MENU_OP_UNFULLSCREEN;
     }
@@ -2715,8 +2627,7 @@ show_window_menu (Client *c, gint px, gint py, guint button, guint32 time)
         ops |= MENU_OP_ABOVE | MENU_OP_BELOW;
     }
 
-    if (clientIsValidTransientOrModal (c) ||
-        FLAG_TEST (c->flags, CLIENT_FLAG_FULLSCREEN))
+    if (is_transient || FLAG_TEST (c->flags, CLIENT_FLAG_FULLSCREEN))
     {
         insensitive |= MENU_OP_NORMAL | MENU_OP_ABOVE | MENU_OP_BELOW;
     }
@@ -2728,7 +2639,7 @@ show_window_menu (Client *c, gint px, gint py, guint button, guint32 time)
         ops |= MENU_OP_CONTEXT_HELP;
     }
 
-    if (clientIsValidTransientOrModal (c)
+    if (is_transient
         || !FLAG_TEST (c->xfwm_flags, XFWM_FLAG_HAS_STICK)
         || FLAG_TEST (c->flags, CLIENT_FLAG_STICKY))
     {
@@ -2839,22 +2750,61 @@ client_event_cb (GtkWidget * widget, GdkEventClient * ev, gpointer data)
     return (FALSE);
 }
 
+static gboolean
+refresh_font_cb (GObject * obj, GdkEvent * ev, gpointer data)
+{
+    ScreenInfo *screen_info;
+
+    screen_info = (ScreenInfo *) data;
+    g_return_val_if_fail (screen_info, TRUE);
+
+    myScreenUpdateFontHeight (screen_info);
+    clientUpdateAllFrames (screen_info, UPDATE_FRAME);
+    return (TRUE);
+}
+
 static void
 size_changed_cb(GdkScreen *gscreen, gpointer data)
 {
     ScreenInfo *screen_info;
     DisplayInfo *display_info;
+    int new_width, new_height;
 
     TRACE ("entering size_changed_cb");
 
     screen_info = (ScreenInfo *) data;
     g_return_if_fail (screen_info);
 
+    /*
+     * We have added/removed a monitor or even changed the layout,
+     * the cache for monitor position we use in our screen structure
+     * is not valid anymore and potentially refers to a monitor that
+     * was just removed, so invalidate it.
+     */
+    screen_info->cache_monitor.x = -1;
+    screen_info->cache_monitor.y = -1;
+    screen_info->cache_monitor.width = 0;
+    screen_info->cache_monitor.height = 0;
+
+    /*
+     * If the overall size of the screen hasn't changed,
+     * there is no need to continue any further...
+     */
+    new_width  = WidthOfScreen (screen_info->xscreen);
+    new_height = HeightOfScreen (screen_info->xscreen);
+
+    if ((screen_info->width  == new_width) &&
+        (screen_info->height == new_height))
+    {
+        return;
+    }
+
     display_info = screen_info->display_info;
-    screen_info->width = WidthOfScreen (screen_info->xscreen);
-    screen_info->height = HeightOfScreen (screen_info->xscreen);
+    screen_info->width = new_width;
+    screen_info->height = new_height;
+
     setNetWorkarea (display_info, screen_info->xroot, screen_info->workspace_count,
-                    screen_info->width, screen_info->height, screen_info->margins);
+                    new_width, new_height, screen_info->margins);
     placeSidewalks (screen_info, screen_info->params->wrap_workspaces);
     clientScreenResize (screen_info);
     compositorUpdateScreenSize (screen_info);
@@ -2863,14 +2813,19 @@ size_changed_cb(GdkScreen *gscreen, gpointer data)
 static void
 monitors_changed_cb(GdkScreen *gscreen, gpointer data)
 {
+    ScreenInfo *screen_info;
+
     TRACE ("entering monitors_changed_cb");
 
+    screen_info = (ScreenInfo *) data;
+    g_return_if_fail (screen_info);
+
     /*
      * From the window manager point of view,
      * a XRand 1.2 monitor change is similar to
      * a screen size change.
      */
-    size_changed_cb(gscreen, data);
+    size_changed_cb (gscreen, data);
 }
 
 void
@@ -2883,15 +2838,15 @@ initGtkCallbacks (ScreenInfo *screen_info)
                           "button_press_event", GTK_SIGNAL_FUNC (show_popup_cb), (gpointer) NULL);
     g_signal_connect (GTK_OBJECT (myScreenGetGtkWidget (screen_info)), "client_event",
                       GTK_SIGNAL_FUNC (client_event_cb), (gpointer) (screen_info->display_info));
-    g_signal_connect(G_OBJECT(screen_info->gscr), "size-changed",
-                     G_CALLBACK(size_changed_cb),
-                     (gpointer) (screen_info));
+    g_signal_connect (G_OBJECT(screen_info->gscr), "size-changed",
+                      G_CALLBACK(size_changed_cb),
+                      (gpointer) (screen_info));
     if(gtk_major_version > 2 || (gtk_major_version == 2 && gtk_minor_version >= 13))
     {
         TRACE ("connect \"monitors-changed\" cb");
-        g_signal_connect(G_OBJECT(screen_info->gscr), "monitors-changed",
-                         G_CALLBACK(monitors_changed_cb),
-                         (gpointer) (screen_info));
+        g_signal_connect (G_OBJECT(screen_info->gscr), "monitors-changed",
+                          G_CALLBACK(monitors_changed_cb),
+                          (gpointer) (screen_info));
     }
 
     settings = gtk_settings_get_default ();
@@ -2903,6 +2858,16 @@ initGtkCallbacks (ScreenInfo *screen_info)
             G_CALLBACK (set_reload), (gpointer) (screen_info->display_info));
         g_signal_connect (settings, "notify::gtk-double-click-time",
             G_CALLBACK (dbl_click_time_cb), (gpointer) (screen_info->display_info));
+        g_signal_connect_after (settings, "notify::gtk-xft-antialias",
+            G_CALLBACK (refresh_font_cb), (gpointer) (screen_info));
+        g_signal_connect_after (settings, "notify::gtk-xft-dpi",
+            G_CALLBACK (refresh_font_cb), (gpointer) (screen_info));
+        g_signal_connect_after (settings, "notify::gtk-xft-hinting",
+            G_CALLBACK (refresh_font_cb), (gpointer) (screen_info));
+        g_signal_connect_after (settings, "notify::gtk-xft-hintstyle",
+            G_CALLBACK (refresh_font_cb), (gpointer) (screen_info));
+        g_signal_connect_after (settings, "notify::gtk-xft-rgba",
+            G_CALLBACK (refresh_font_cb), (gpointer) (screen_info));
     }
 }
 
